// [v0.1.046-20250329]

// === module init block ===

const {
  tryToNumber,
  isArray, isPlainObject, isDateObject,
} = require('@ygracs/bsfoc-lib-js');

// === module extra block (helper functions) ===

/**
 * @function isValidSeparator
 * @param {string} value
 * @param {string[]} list
 * @returns {boolean}
 * @inner
 * @description Checks whether a given value match a given list
 *              of a separators.
 */
function isValidSeparator(value, list) {
  const sample = typeof value === `string` ? value.trim() : '';
  return (
    isArray(list)
    && sample !== ''
    && list.indexOf(sample) !== -1
  );
};

/**
 * @function __dateLocalToStr
 * @param {Date} dt - a value to convert
 * @param {string} [sep=""] - a separator
 * @returns {string}
 * @inner
 * @description Returns a local date as a string in format `YYYY<sep>MM<sep>DD`.
 */
function __dateLocalToStr(dt, sep = '') {
  return [
    dt.getFullYear().toString().padStart(4, "0"),
    (dt.getMonth() + 1).toString().padStart(2, "0"),
    dt.getDate().toString().padStart(2, "0"),
  ].join(sep);
};

/**
 * @function __dateUTCToStr
 * @param {Date} dt - a value to convert
 * @param {string} [sep=""] - a separator
 * @returns {string}
 * @inner
 * @description Returns a date in UTC as a string in format `YYYY<sep>MM<sep>DD`.
 */
function __dateUTCToStr(dt, sep = '') {
  return [
    dt.getUTCFullYear().toString().padStart(4, "0"),
    (dt.getUTCMonth() + 1).toString().padStart(2, "0"),
    dt.getUTCDate().toString().padStart(2, "0"),
  ].join(sep);
};

/**
 * @function __timeLocalToStr
 * @param {Date} dt - a value to convert
 * @param {string} [sep=""] - a separator
 * @returns {string}
 * @inner
 * @description Returns a local time as a string in format `hh<sep>mm<sep>ss`.
 */
function __timeLocalToStr(dt, sep = '') {
  return [
    dt.getHours().toString().padStart(2, "0"),
    dt.getMinutes().toString().padStart(2, "0"),
    dt.getSeconds().toString().padStart(2, "0"),
  ].join(sep);
};

/**
 * @function __timeUTCToStr
 * @param {Date} dt - a value to convert
 * @param {string} [sep=""] - a separator
 * @returns {string}
 * @inner
 * @description Returns a time in UTC as a string in format `hh<sep>mm<sep>ss`.
 */
function __timeUTCToStr(dt, sep = '') {
  return [
    dt.getUTCHours().toString().padStart(2, "0"),
    dt.getUTCMinutes().toString().padStart(2, "0"),
    dt.getUTCSeconds().toString().padStart(2, "0"),
  ].join(sep);
};

/**
 * @typedef {Object} OPT_PS_TimeToStr
 * @property {boolean} addSec - flag that require to add seconds
 * @property {boolean} addMSec - flag that require to add milliseconds
 * @property {string} timePartSep - time parts separator
 * @property {string} msecPartSep - milliseconds part separator
 * @description An object that contains settings used for converting
 *              a time part of a `DateTime`-object into its string
 *              representation.
 */

/**
 * @function __timeLocalToStrEx
 * @param {Date} dt - a value to convert
 * @param {OPT_PS_TimeToStr} [opt] - options
 * @returns {string}
 * @inner
 * @description Returns a local time as a string in format `hh:mm:ss`.
 */
function __timeLocalToStrEx(dt, opt) {
  /** @type {OPT_PS_TimeToStr} */
  let {
    addSec,
    addMSec,
    timePartSep: sep,
    msecPartSep: msep,
  } = opt;
  if (typeof addSec !== 'boolean') addSec = true;
  if (typeof addMSec !== 'boolean') addMSec = false;
  sep = typeof sep === 'string' ? sep.trim() : '';
  let result = [
    dt.getHours().toString().padStart(2, "0"),
    dt.getMinutes().toString().padStart(2, "0"),
  ].join(sep);
  if (addSec) {
    result += `${sep}${dt.getSeconds().toString().padStart(2, "0")}`;
    if (addMSec && typeof msep === 'string') {
      msep = msep.trim();
      if (msep === '.' || msep === ',') {
        result += `${msep}${dt.getMilliseconds().toString()}`;
      };
    };
  };
  return result;
};

/**
 * @function __timeUTCToStrEx
 * @param {Date} dt - a value to convert
 * @param {OPT_PS_TimeToStr} [opt] - options
 * @returns {string}
 * @inner
 * @description Returns a time in UTC as a string in format `hh:mm:ss`.
 */
function __timeUTCToStrEx(dt, opt) {
  /** @type {OPT_PS_TimeToStr} */
  let {
    addSec,
    addMSec,
    timePartSep: sep,
    msecPartSep: msep,
  } = opt;
  if (typeof addSec !== 'boolean') addSec = true;
  if (typeof addMSec !== 'boolean') addMSec = false;
  sep = typeof sep === 'string' ? sep.trim() : '';
  let result = [
    dt.getUTCHours().toString().padStart(2, "0"),
    dt.getUTCMinutes().toString().padStart(2, "0"),
  ].join(sep);
  if (addSec) {
    result += `${sep}${dt.getUTCSeconds().toString().padStart(2, "0")}`;
    if (addMSec && typeof msep === 'string') {
      msep = msep.trim();
      if (msep === '.' || msep === ',') {
        result += `${msep}${dt.getUTCMilliseconds().toString()}`;
      };
    };
  };
  return result;
};

/**
 * @typedef {Object} OPT_PS_DateToStr
 * @property {string} datePartSep - date part separator
 */

/**
 * @function __getDateAsString
 * @param {Date} dt - a value to convert
 * @param {OPT_PS_DateToStr} [opt] - options
 * @returns {string}
 * @inner
 * @description Converts a `DateTime` to string.
 */
function __getDateAsString(dt, opt) {
  /** @type {OPT_PS_DateToStr} */
  let {
    utc_src,
    datePartSep: sep,
  } = opt;
  if (typeof utc_src !== 'boolean') utc_src = false;
  sep = trySeparator(sep, ALLOWED_DATE_SEP_LIST);
  return (
    utc_src
    ? __dateUTCToStr(dt, sep)
    : __dateLocalToStr(dt, sep)
  );
};

/**
 * @function __convValueToTZString
 * @param {number} value - a value to convert
 * @param {string} [sep=""] - a separator
 * @returns {string}
 * @inner
 * @description Converts a `TZ` from a number to a string
 *              in format (-|+)hh<sep>mm.
 */
function __convValueToTZString(value, sep = '') {
  const tzH = Math.trunc(value / 60);
  const tzM = value - (tzH * 60);
  // // TODO: check TZ range
  return [
    (value < 0) ? '-' : '+',
    Math.abs(tzH).toString().padStart(2, "0"),
    sep,
    Math.abs(tzM).toString().padStart(2, "0"),
  ].join('');
};

// === module main block ===

/***
 * (* constant definitions *)
 */

/**
 * defines a list of allowed separators for date elements
 * @inner
 */
const ALLOWED_DATE_SEP_LIST = [ '.', '-', '/' ];

/***
 * (* function definitions *)
 */

/**
 * @function trySeparator
 * @param {string} value - a value to check
 * @param {string[]} list - an array of allowed separators
 * @returns {?string}
 * @since v0.0.26
 * @description Evaluate a given value as a separator.
 */
function trySeparator(value, list) {
  let result = null;
  if (isArray(list)) {
    let sample = value;
    if (typeof sample !== 'string') {
      sample = list[0];
    } else {
      if (
        sample !== ''
        && ((sample = sample.trim()) !== '')
        && list.indexOf(sample) === -1
      ) {
        sample = list[0];
      };
    };
    if (typeof sample === 'string') result = sample
  };
  return result;
};

/**
 * @function getDateAsString
 * @param {(number|string|Date)} date - a timestamp or DateTime-object
 * @param {OPT_PS_DateToStr} [opt] - options
 * @returns {string}
 * @description Converts a `DateTime` to string.
 */
function getDateAsString(date, opt) {
  const { isSucceed, value: dtValue } = tryDTValue(date);
  if (!isSucceed) return '';
  const _opt = isPlainObject(opt) ? opt : {};
  return __getDateAsString(dtValue, _opt);
};

/**
 * @typedef {Object} OPT_PS_DTvToISO
 * @property {boolean} [addDate=true] - flag that require to add date
 * @property {boolean} [useBaseFormat=false] - flag that require to use a base format
 * @description An object that contains settings used for converting
 *              a `DateTime`-object into its string representation in ISO8601-format.
 */

/**
 * @function convDTValueToISO8601
 * @param {(number|string|Date)} date - a timestamp or DateTime-object
 * @param {OPT_PS_DTvToISO} [opt] - options
 * @returns {string}
 * @description Converts a `DateTime` to a string formated in ISO8601-format.
 */
function convDTValueToISO8601(date, opt) {
  /** @type {OPT_PS_DTvToISO} */
  let {
    addDate,
    useBaseFormat: useBsF,
  } = isPlainObject(opt) ? opt : {};
  if (typeof useBsF !== 'boolean') useBsF = false;
  if (typeof addDate !== 'boolean') addDate = true;
  const { isSucceed, value: dtValue } = tryDTValue(date);
  if (!isSucceed) return '';
  const dSep = useBsF ? '' : '-';
  const tSep = useBsF ? '' : ':';
  const tz = dtValue.getTimezoneOffset();
  return [
    addDate ? __dateLocalToStr(dtValue, dSep) : '',
    'T',
    __timeLocalToStr(dtValue,tSep),
    __convValueToTZString(tz * -1, tSep),
  ].join('');
};

/**
 * @function convDTValueToXMLTV
 * @param {(number|string|Date)} date - a timestamp or DateTime-object
 * @param {object} [opt] - options
 * @returns {string}
 * @description Converts a `DateTime` to a string formated in XMLTV-format.
 */
function convDTValueToXMLTV(date, opt) {
  let {
    src_in_tz: dtShft,
  } = isPlainObject(opt) ? opt : {};
  let { isSucceed, value: dtValue } = tryDTValue(date);
  if (!isSucceed) return '';
  /**/// // TODO: consider the use of 'dtShft' option
  //if (dtShft !== 0) dtValue.setUTCMinutes(dtValue.getUTCMinutes() - dtShft);
  return [
    __dateLocalToStr(dtValue, ''),
    __timeLocalToStr(dtValue, ''),
    ' ',
    __convValueToTZString(dtValue.getTimezoneOffset() * -1, ''),
  ].join('');
};

/**
 * @function convXMLTVToDTValue
 * @param {string} value
 * @param {boolean} [opt=false] - options
 * @returns {(number|Date)}
 * @description Converts a string formated in XMLTV-format to a `DateTime`.
 */
function convXMLTVToDTValue(value, opt) {
  const doRetTS = typeof opt === 'boolean' ? opt : false;
  let result = null;
  let _value = '';
  if (
    typeof value === 'string' && ((_value = value.trim()) !== '')
  ) {
    let dtShft = 0;
    let [ dtStr, tzStr ] = _value.split(' ');
    //console.log('CHECK: dtStr => ['+dtStr+']');
    //console.log('CHECK: tzStr => ['+tzStr+']');
    if (!Number.isNaN(Number(dtStr))) {
      if (!Number.isNaN(Number(tzStr))) {
        let value = [
          dtStr.substring(0, 4),
          '-',
          dtStr.substring(4, 6),
          '-',
          dtStr.substring(6, 8),
          'T',
          dtStr.substring(8, 10),
          ':',
          dtStr.substring(10, 12),
          ':',
          dtStr.substring(12, 14),
          '.000',
          tzStr.substring(0, 3),
          ':',
          tzStr.substring(3, 5),
        ].join('');
        //console.log('CHECK: value => ['+value+']');
        let dtValue = new Date(value);
        result = doRetTS ? dtValue.getTime() : dtValue;
      };
    };
  };
  return result;
};

/**
 * @typedef {Object} OPT_PS_DTvToStr
 * @property {boolean} addDate - flag that require to add date
 * @property {boolean} addTime - flag that require to add time
 * @property {string} datePartSep - date part separator
 * @property {boolean} addSec - flag that require to add seconds
 * @property {boolean} addMSec - flag that require to add milliseconds
 * @property {string} timePartSep - time parts separator
 * @property {string} msecPartSep - milliseconds part separator
 * @description An object that contains settings used for converting
 *              a `DateTime`-object into its string representation.
 */

/**
 * @function convDTValueToDateString
 * @param {(number|string|Date)} date - a timestamp or DateTime-object
 * @param {OPT_PS_DTvToStr} [opt] - options
 * @returns {string}
 * @description Converts a `DateTime` to a formated string.
 */
function convDTValueToDateString(date, opt) {
  let { isSucceed, value: dtValue } = tryDTValue(date);
  if (!isSucceed) return '';
  let result = '';
  /** @type {OPT_PS_DTvToStr} */
  const _opt = isPlainObject(opt) ? opt : {};
  let {
    addDate,
    addTime,
    src_in_utc: utc_src,
  } = _opt;
  if (typeof utc_src !== 'boolean') utc_src = false;
  if (typeof addDate !== 'boolean') addDate = true;
  if (typeof addTime !== 'boolean') addTime = true;
  if (addDate || addTime) {
    if (addDate) result = __getDateAsString(dtValue, _opt);
    if (addDate && addTime) result += ' ';
    if (addTime) {
      let {
        addSec,
        addMSec, // <* reserved *>
        timePartSep, // <* reserved *>
        msecPartSep, // <* reserved *>
      } = _opt;
      result += (
        utc_src
        ? __timeUTCToStrEx(dtValue, { addSec, addMSec, timePartSep: ':', msecPartSep })
        : __timeLocalToStrEx(dtValue, { addSec, addMSec, timePartSep: ':', msecPartSep })
      );
    };
  };
  // TODO: check TZ
  //console.log(result);
  return result;
};

/**
 * @function convTZStringToValue
 * @param {string} value
 * @returns {number}
 * @description Converts a `TZ` from a string to a number.
 */
function convTZStringToValue(value) {
  let result = NaN;
  if (typeof value === 'string') {
    let tzValue = value.trim();
    if (tzValue !== '') {
      let te = tzValue.match(/[+-](\d\d):?(\d\d)?(.)?/);
      //console.log(te);
      if (te !== null && te[3] === undefined) {
        let [ c_str, tzH, tzM ] = te;
        if (tzM === undefined) tzM = 0;
        result = Number(tzH) * 60 + Number(tzM);
        if (result !== 0 && c_str.charAt(0) === '-') result *= -1;
      };
    };
  };
  return result;
};

/**
 * @function convValueToTZString
 * @param {(number|string)} value - some value
 * @param {string} [sep=''] - <reserved>
 * @returns {string}
 * @description Converts a `TZ` from a number to a string.
 */
function convValueToTZString(value, sep) {
  const tzValue = tryToNumber(value);
  if (tzValue === null) return '';
  let _sep = typeof sep === 'string' ? sep.trim() : '';
  if (sep !== '' || sep !== ':') sep = '';
  return __convValueToTZString (tzValue, _sep);
};

/**
 * @typedef {Object} VCOR_trydt
 * @property {boolean} isSucceed - ops flag
 * @property {?Date} value - result value
 * @description A result of a value check ops.
 */

/**
 * @function tryDTValue
 * @param {any} value - some value
 * @returns {VCOR_trydt}
 * @description Tries to convert a given value to `DateTime`.
 */
function tryDTValue(value) {
  let isSucceed = false;
  let result = NaN;
  if (isDateObject(value)) {
    result = value;
    isSucceed = true;
  } else {
    const _value = tryToNumber(value);
    if (_value !== null) {
      result = new Date(_value);
      isSucceed = true;
    };
  };
  return {
    isSucceed,
    value: result,
  };
};

/**
 * @function timeLocalToString
 * @param {(number|string|Date)} date - a timestamp or DateTime-object
 * @param {OPT_PS_TimeToStr} [opt] - options
 * @returns {string}
 * @description Returns a local time as a string in format `hh:mm:ss`.
 */
function timeLocalToString(date, opt) {
  const { isSucceed, value: dt } = tryDTValue(date);
  if (!isSucceed) return '';
  /** @type {OPT_PS_TimeToStr} */
  let {
    addSec,
    addMSec,
    timePartSep: sep, // <*reserved*>
    msecPartSep, // <*reserved*>
  } = isPlainObject(opt) ? opt : {};
  if (typeof sep !== 'string') sep = ':';
  return __timeLocalToStrEx(dt, { addSec, addMSec, timePartSep: sep, msecPartSep });
};

/**
 * @function timeUTCToString
 * @param {(number|string|Date)} date - a timestamp or DateTime-object
 * @param {OPT_PS_TimeToStr} [opt] - options
 * @returns {string}
 * @description Returns a time in UTC as a string in format `hh:mm:ss`.
 */
function timeUTCToString(date, opt) {
  const { isSucceed, value: dt } = tryDTValue(date);
  if (!isSucceed) return '';
  /** @type {OPT_PS_TimeToStr} */
  let {
    addSec,
    addMSec,
    timePartSep: sep, // <*reserved*>
    msecPartSep, // <*reserved*>
  } = isPlainObject(opt) ? opt : {};
  if (typeof sep !== 'string') sep = ':';
  return __timeUTCToStrEx(dt, { addSec, addMSec, timePartSep: sep, msecPartSep });
};

/***
 * (* class definitions *)
 */

// === module exports block ===

module.exports.isDateObject = isDateObject;
module.exports.convDTValueToISO8601 = convDTValueToISO8601;
module.exports.convDTValueToXMLTV = convDTValueToXMLTV;
module.exports.convXMLTVToDTValue = convXMLTVToDTValue;
module.exports.convDTValueToDateString = convDTValueToDateString;
module.exports.convValueToTZString = convValueToTZString;
module.exports.convTZStringToValue = convTZStringToValue;
module.exports.getDateAsString = getDateAsString;
module.exports.timeLocalToString = timeLocalToString;
module.exports.timeUTCToString = timeUTCToString;
module.exports.tryDTValue = tryDTValue;
module.exports.trySeparator = trySeparator;
